from enum import Enum


class RunMethod(Enum):
    """Plugin Execution Methods."""

    JOB = 'job'
    TASK = 'task'
    RESTAPI = 'restapi'


class PluginCategory(Enum):
    NEURAL_NET = 'neural_net'
    EXPORT = 'export'
    UPLOAD = 'upload'
    SMART_TOOL = 'smart_tool'
    POST_ANNOTATION = 'post_annotation'
    PRE_ANNOTATION = 'pre_annotation'
    DATA_VALIDATION = 'data_validation'

    @classmethod
    def choices(cls):
        return [(member.value, member.name.replace('_', ' ').title()) for member in cls]



import os
import sys

import click
from dotenv import load_dotenv

from synapse_sdk.cli.alias.utils import load_dotenv_default_alias
from synapse_sdk.i18n import gettext as _

from .create import create
from .publish import publish
from .run import run

load_dotenv_default_alias()
load_dotenv(os.path.join(os.getcwd(), '.env'), override=True)
sys.path.append(os.getcwd())


@click.group(context_settings={'obj': {}, 'auto_envvar_prefix': 'SYNAPSE_PLUGIN'})
@click.option('--debug/--no-debug', default=False)
@click.pass_context
def plugin(ctx, debug):
    """Manage Synapse plugins."""
    ctx.ensure_object(dict)
    ctx.obj['DEBUG'] = debug

    if debug:
        click.echo(_('Debug mode is "on"'))


plugin.add_command(create)
plugin.add_command(publish)
plugin.add_command(run)



import subprocess
import tempfile
from pathlib import Path

from synapse_sdk.i18n import gettext as _
from synapse_sdk.utils.file import calculate_checksum, download_file
from synapse_sdk.utils.storage import get_storage


def archive(source_path, archive_path):
    archive_path.parent.mkdir(parents=True, exist_ok=True)
    command = f'git ls-files --others --exclude-standard --cached  | zip -q --names-stdin {archive_path}'
    subprocess.run(command, cwd=source_path, shell=True, check=True, stdout=subprocess.DEVNULL)


def download_and_upload(source_url, url):
    storage = get_storage(url)
    with tempfile.TemporaryDirectory() as temp_path:
        file_path = str(download_file(source_url, temp_path))
        checksum = calculate_checksum(file_path)
        # TODO 중복 체크
        return storage.upload(file_path, f'dev-{checksum}.zip')


def archive_and_upload(source_path, url):
    storage = get_storage(url)
    dist_path = Path(source_path, 'dist')
    archive_path = dist_path / 'archive.zip'

    archive(source_path, archive_path)
    checksum = calculate_checksum(archive_path)
    checksum_archive_path = dist_path / f'dev-{checksum}.zip'

    if checksum_archive_path.exists():
        # TODO 실제 스토리지 있는지 확인
        return storage.get_url(checksum_archive_path.name)

    archive_path.rename(checksum_archive_path)
    for file_path in dist_path.glob('*.zip'):
        if file_path.name != checksum_archive_path.name:
            file_path.unlink()
    return storage.upload(str(checksum_archive_path), checksum_archive_path.name)


def build_and_upload(source_path, url, virtualenv_path='.venv'):
    storage = get_storage(url)
    dist_path = Path(source_path, 'dist')
    archive_path = dist_path / 'archive.zip'

    archive(source_path, archive_path)
    checksum = calculate_checksum(archive_path)
    checksum_archive_path = dist_path / f'dev-{checksum}.zip'

    if checksum_archive_path.exists():
        # TODO 실제 스토리지 있는지 확인
        wheel_path = next(dist_path.glob('*.whl'), None)
        return storage.get_url(wheel_path.name)

    # wheel file 빌드 진행
    for file_path in dist_path.glob('*.whl'):
        file_path.unlink()

    print(_('Building {}...').format(Path(source_path).name))
    subprocess.run(
        f'{virtualenv_path}/bin/python -m build --wheel',
        cwd=source_path,
        shell=True,
        check=True,
        stdout=subprocess.DEVNULL,
    )
    wheel_path = next(dist_path.glob('*.whl'), None)

    # whl 파일 버전이 동일한 이슈를 해결하기 위해 checksum으로 build명 변경
    checksum_wheel_path = wheel_path.with_name(change_build_from_whl_name(wheel_path.name, checksum))
    wheel_path.rename(checksum_wheel_path)

    archive_path.rename(checksum_archive_path)

    for file_path in dist_path.glob('*.zip'):
        if file_path.name != checksum_archive_path.name:
            file_path.unlink()
    return storage.upload(str(checksum_wheel_path), checksum_wheel_path.name)


def change_build_from_whl_name(whl_name, build):
    components = whl_name.split('-')
    version = components[1].split('+')[0]
    components[1] = f'{version}+{build}'
    return '-'.join(components)



from upath import UPath

from synapse_sdk.utils.storage.providers import BaseStorage


class SFTPStorage(BaseStorage):
    def get_pathlib(self, path):
        credentials = self.query_params['params']
        host = self.query_params['host']
        root_path = self.query_params['root_path']

        username = credentials['username']
        password = credentials['password']
        if path == '/':
            path = ''
        return UPath(f'sftp://{host}', username=username, password=password) / root_path / path



from urllib.parse import urlparse, urlunparse


def clean_url(url, remove_query_params=True, remove_fragment=True):
    parsed = urlparse(url)
    query = '' if remove_query_params else parsed.query
    fragment = '' if remove_fragment else parsed.fragment

    return urlunparse((
        parsed.scheme,
        parsed.netloc,
        parsed.path,
        parsed.params,
        query,
        fragment,
    ))



import json
from pathlib import Path

import requests


def export(run, export_items, path_root, **params):
    """Executes the export task.

    Args:
        run : Execution object
        export_items (generator):
            - data (dict): dm_schema_data information.
            - files (dict): File information. Includes file URL, original file path, metadata, etc.
            - id (int): ground_truth ID
        path_root : pathlib object, the path to export
        **params: Additional parameters

    Returns:
        dict: Result
    """

    path_root.mkdir(parents=True, exist_ok=True)
    run.log_message('Starting export process.')

    # results contains all information fetched through the list API.
    # example:
    #   params.get('results', [])

    save_original_file_flag = params.get('save_original_file')
    errors_json_file_list = []
    errors_original_file_list = []

    # Path to save JSON files
    json_output_path = path_root / 'json'
    json_output_path.mkdir(parents=True, exist_ok=True)

    # Path to save original files
    origin_files_output_path = path_root / 'origin_files'
    origin_files_output_path.mkdir(parents=True, exist_ok=True)

    total = params['count']
    # progress init
    run.set_progress(0, total, category='dataset_conversion')
    for no, export_item in enumerate(export_items, start=1):
        run.set_progress(no, total, category='dataset_conversion')
        if no == 1:
            run.log_message('Converting dataset.')
        preprocessed_data = before_convert(export_item)
        converted_data = convert_data(preprocessed_data)
        final_data = after_convert(converted_data)

        # Call if original file extraction is needed
        if save_original_file_flag:
            if no == 1:
                run.log_message('Saving original file.')
            save_original_file(final_data, origin_files_output_path, errors_original_file_list)

        # Extract data as JSON files
        if no == 1:
            run.log_message('Saving json file.')
        save_as_json(final_data, json_output_path, errors_json_file_list)

    run.end_log()

    # Save error list files
    if len(errors_json_file_list) > 0 or len(errors_original_file_list) > 0:
        export_error_file = {'json_file_name': errors_json_file_list, 'origin_file_name': errors_original_file_list}
        with (path_root / 'error_file_list.json').open('w', encoding='utf-8') as f:
            json.dump(export_error_file, f, indent=4, ensure_ascii=False)

    return {'export_path': str(path_root)}


def convert_data(data):
    """Converts the data."""
    return data


def before_convert(data):
    """Preprocesses the data before conversion."""
    return data


def after_convert(data):
    """Post-processes the data after conversion."""
    return data


def get_original_file_pathlib(files):
    """Retrieve the original file path from the given file information.

    Args:
        files (dict): A dictionary containing file information, including file URL,
                      original file path, metadata, etc.

    Returns:
        pathlib.Path: The original file path extracted from the metadata.
    """
    return Path(files['meta']['path_original'])


def save_original_file(result, base_path, error_file_list):
    """Saves the original file.

    Args:
        result (dict): API response data containing file information.
        base_path (Path): The directory where the file will be saved.
        error_file_list (list): A list to store error files.
    """
    file_url = result['files']['url']
    file_name = get_original_file_pathlib(result['files']).name
    response = requests.get(file_url)
    try:
        with (base_path / file_name).open('wb') as file:
            file.write(response.content)
    except Exception as e:
        error_file_list.append([file_name, str(e)])


def save_as_json(result, base_path, error_file_list):
    """Saves the data as a JSON file.

    Args:
        result (dict): API response data containing file information.
        base_path (Path): The directory where the file will be saved.
        error_file_list (list): A list to store error files.
    """
    # Default save file name: original file name
    file_name = get_original_file_pathlib(result['files']).stem
    json_data = result['data']
    try:
        with (base_path / f'{file_name}.json').open('w', encoding='utf-8') as f:
            json.dump(json_data, f, indent=4, ensure_ascii=False)
    except Exception as e:
        error_file_list.append([f'{file_name}.json', str(e)])



from synapse_sdk.clients.exceptions import ClientError
from synapse_sdk.plugins.categories.base import Action
from synapse_sdk.plugins.categories.decorators import register_action
from synapse_sdk.plugins.enums import PluginCategory, RunMethod


@register_action
class DeploymentAction(Action):
    name = 'deployment'
    category = PluginCategory.NEURAL_NET
    method = RunMethod.JOB

    def get_actor_options(self):
        options = {'runtime_env': self.get_runtime_env()}
        for option in ['num_cpus', 'num_gpus']:
            option_value = self.params.get(option)
            if option_value:
                options[option] = option_value
        return options

    def start(self):
        from ray import serve

        from synapse_sdk.plugins.categories.neural_net.base.inference import app

        self.ray_init()

        deployment = serve.deployment(ray_actor_options=self.get_actor_options())(serve.ingress(app)(self.entrypoint))
        serve.delete(self.plugin_release.code)

        # TODO add run object
        serve.run(
            deployment.bind(self.envs['SYNAPSE_PLUGIN_RUN_HOST']),
            name=self.plugin_release.code,
            route_prefix=f'/{self.plugin_release.checksum}',
        )

        # 백엔드에 ServeApplication 추가
        serve_application = self.create_serve_application()
        return {'serve_application': serve_application['id'] if serve_application else None}

    def create_serve_application(self):
        if self.job_id:
            try:
                serve_application = self.ray_client.get_serve_application(self.plugin_release.code)
                return self.client.create_serve_application({
                    'job': self.job_id,
                    'status': serve_application['status'],
                    'data': serve_application,
                })
            except ClientError:
                pass
        return None



from synapse_sdk.plugins.categories.base import Action
from synapse_sdk.plugins.categories.registry import _REGISTERED_ACTIONS


def register_action(action_class):
    if not issubclass(action_class, Action):
        raise ValueError('Wrapped class must subclass Action class.')

    try:
        _REGISTERED_ACTIONS[action_class.category.value][action_class.name] = action_class
    except KeyError:
        _REGISTERED_ACTIONS[action_class.category.value] = {action_class.name: action_class}
    return action_class



import click

from .create import create
from .default import default
from .delete import delete
from .list import list as list_cli
from .read import read
from .update import update


@click.group()
def alias():
    """Manage aliases"""
    pass


alias.add_command(list_cli)
alias.add_command(create)
alias.add_command(read)
alias.add_command(update)
alias.add_command(delete)
alias.add_command(default)



from synapse_sdk.clients.base import BaseClient


class ServiceClientMixin(BaseClient):
    def run_plugin_release(self, code, data):
        path = f'plugin_releases/{code}/run/'
        return self._post(path, data=data)

    def run_debug_plugin_release(self, data):
        path = 'plugin_releases/run_debug/'
        return self._post(path, data=data)

    def create_plugin_release(self, data):
        path = 'plugin_releases/'
        return self._post(path, data=data)



from synapse_sdk.plugins.categories.base import Action
from synapse_sdk.plugins.categories.decorators import register_action
from synapse_sdk.plugins.enums import PluginCategory, RunMethod


@register_action
class PostAnnotationAction(Action):
    name = 'post_annotation'
    category = PluginCategory.POST_ANNOTATION
    method = RunMethod.TASK



from enum import Enum
from typing import Dict

from pydantic import BaseModel


class StorageCategory(str, Enum):
    """Synapse Backend Storage Category Enum."""

    INTERNAL = 'internal'
    EXTERNAL = 'external'


class StorageProvider(str, Enum):
    """Synapse Backend Storage Provider Enum."""

    AMAZON_S3 = 'amazon_s3'
    AZURE = 'azure'
    DIGITAL_OCEAN = 'digital_ocean'
    FILE_SYSTEM = 'file_system'
    FTP = 'ftp'
    SFTP = 'sftp'
    MINIO = 'minio'
    GCP = 'gcp'


class Storage(BaseModel):
    """Synapse Backend Storage Model.

    Attrs:
        id (int): The storage pk.
        name (str): The storage name.
        category (str): The storage category. (ex: internal, external)
        provider (str): The storage provider. (ex: s3, gcp)
        configuration (Dict): The storage configuration.
        is_default (bool): The storage is default for Synapse backend workspace.
    """

    id: int
    name: str
    category: StorageCategory
    provider: StorageProvider
    configuration: Dict
    is_default: bool



from pathlib import Path

import click

from synapse_sdk.clients.backend import BackendClient
from synapse_sdk.i18n import gettext as _
from synapse_sdk.plugins.models import PluginRelease
from synapse_sdk.plugins.upload import archive


@click.command()
@click.option('--host', required=True)
@click.option('--user_token', required=True)
@click.option('--tenant', required=True)
@click.option('--debug_modules', default='', envvar='SYNAPSE_DEBUG_MODULES')
@click.pass_context
def publish(ctx, host, user_token, tenant, debug_modules):
    debug = ctx.obj['DEBUG']

    plugin_release = PluginRelease()

    source_path = Path('./')
    archive_path = source_path / 'dist' / 'archive.zip'
    archive(source_path, archive_path)

    data = {'plugin': plugin_release.plugin, 'file': str(archive_path), 'debug': debug}
    if debug:
        modules = debug_modules.split(',') if debug_modules else []
        data['meta'] = {'modules': modules}

    client = BackendClient(host, user_token, tenant=tenant)
    client.create_plugin_release(data)
    click.secho(
        _('Successfully published "{}" ({}) to synapse backend!').format(plugin_release.name, plugin_release.code),
        fg='green',
        bold=True,
    )



from ray import serve


class MockNetInference:
    model_id = None

    @serve.multiplexed()
    async def get_model(self, model_id: str):
        return model_id

    async def __call__(self):
        model_id = await self.get_model(serve.get_multiplexed_model_id())
        message = 'hello datamaker!!!'
        return {'model_id': model_id, 'message': message}



from typing import Any, Dict

from synapse_sdk.utils.pydantic.config import ERROR_MESSAGES


def pydantic_to_drf_error(e):
    """
    Convert a pydantic ValidationError into a DRF-style error response.
    """
    drf_errors: Dict[str, Any] = {}

    for error in e.errors():
        field_path = error['loc']
        context_error = error.get('ctx', {}).get('error')

        error_msg = context_error or ERROR_MESSAGES.get(error['type'], error['msg'])

        # Convert the field path into a nested dictionary structure
        current = drf_errors
        for i, key in enumerate(field_path[:-1]):
            current = current.setdefault(str(key), {})

        # Set the error message at the final location
        final_key = str(field_path[-1])
        if final_key in current:
            if isinstance(current[final_key], list):
                current[final_key].append(error_msg)
            else:
                current[final_key] = [current[final_key], error_msg]
        else:
            current[final_key] = [error_msg]

    return drf_errors



import inspect
import json
import os
from functools import cached_property
from pprint import pprint

import requests
from pydantic import ValidationError

from synapse_sdk.clients.ray import RayClient
from synapse_sdk.plugins.enums import RunMethod
from synapse_sdk.plugins.exceptions import ActionError
from synapse_sdk.plugins.models import PluginRelease, Run
from synapse_sdk.plugins.upload import archive_and_upload, build_and_upload, download_and_upload
from synapse_sdk.utils.module_loading import import_string
from synapse_sdk.utils.pydantic.errors import pydantic_to_drf_error


class Action:
    """Base class for all plugin actions.

    Attrs:
        name (str): The name of the action.
        category (PluginCategory): The category of the action.
        method (RunMethod): The method to run of the action.
        run_class (Run): The class to run the action.
        params_model (BaseModel): The model to validate the params.
        progress_categories (List[str]): The categories to update the progress.
        params (Dict): The params to run the action.
        plugin_config (Dict): The plugin config.
        plugin_release (PluginRelease): The plugin release.
        config (Dict): The action config.
        requirements (List[str]): The requirements to install.
        job_id (str): The job id.
        direct (bool): The flag to run the action directly.
        debug (bool): The flag to run the action in debug mode.
        envs (Dict): The runtime envs.
        run (Run): The run instance.

    Raises:
        ActionError: If the action fails.
    """

    # class 변수
    name = None
    category = None
    method = None
    run_class = Run
    params_model = None
    progress_categories = None

    # init 변수
    params = None
    plugin_config = None
    plugin_release = None
    config = None
    requirements = None
    job_id = None
    direct = None
    debug = None
    envs = None
    run = None

    REQUIRED_ENVS = [
        'RAY_ADDRESS',
        'RAY_DASHBOARD_URL',
        'RAY_SERVE_ADDRESS',
        'SYNAPSE_PLUGIN_STORAGE',
        'SYNAPSE_DEBUG_PLUGIN_PATH',
        'SYNAPSE_DEBUG_MODULES',
        'SYNAPSE_PLUGIN_RUN_HOST',
        'SYNAPSE_PLUGIN_RUN_USER_TOKEN',
        'SYNAPSE_PLUGIN_RUN_TENANT',
    ]

    def __init__(self, params, plugin_config, requirements=None, envs=None, job_id=None, direct=False, debug=False):
        self.params = params
        self.plugin_config = plugin_config
        self.plugin_release = PluginRelease(config=plugin_config)
        self.config = self.plugin_release.get_action_config(self.name)
        self.requirements = requirements
        self.job_id = job_id
        self.direct = direct
        self.debug = debug
        self.envs = {**self.get_default_envs(), **envs} if envs else self.get_default_envs()
        self.run = self.get_run()

    @cached_property
    def entrypoint(self):
        return import_string(self.config['entrypoint'])

    @property
    def plugin_storage_url(self):
        return self.envs['SYNAPSE_PLUGIN_STORAGE']

    @property
    def client(self):
        return self.run.client

    @property
    def ray_client(self):
        return RayClient(self.envs['RAY_DASHBOARD_URL'])

    @property
    def plugin_url(self):
        if self.debug:
            plugin_path = self.envs.get('SYNAPSE_DEBUG_PLUGIN_PATH') or '.'
            if plugin_path.startswith('https://'):  # TODO ray에서 지원하는 remote uri 형식 (https, s3, gs) 모두 지원
                plugin_url = plugin_path
            elif plugin_path.startswith('http://'):
                plugin_url = download_and_upload(plugin_path, self.plugin_storage_url)
            else:
                plugin_url = archive_and_upload(plugin_path, self.plugin_storage_url)
            self.envs['SYNAPSE_DEBUG_PLUGIN_PATH'] = plugin_url
            return plugin_url
        return self.plugin_release.get_url(self.plugin_storage_url)

    @property
    def debug_modules(self):
        debug_modules = []
        if self.envs.get('SYNAPSE_DEBUG_MODULES'):
            for module_path in self.envs['SYNAPSE_DEBUG_MODULES'].split(','):
                # TODO ray에서 지원하는 remote uri 형식 (https, s3, gs) 모두 지원
                if module_path.startswith('https://'):
                    module_url = module_path
                else:
                    module_url = build_and_upload(module_path, self.plugin_storage_url)
                debug_modules.append(module_url)
            self.envs['SYNAPSE_DEBUG_MODULES'] = ','.join(debug_modules)
        return debug_modules

    @property
    def plugin_package_manager(self):
        return self.plugin_config.get('package_manager', 'pip')

    def get_run(self):
        context = {
            'plugin_release': self.plugin_release,
            'progress_categories': self.progress_categories,
            'params': self.params,
            'envs': self.envs,
            'debug': self.debug,
        }
        return self.run_class(self.job_id, context)

    def get_default_envs(self):
        return {env: os.environ[env] for env in self.REQUIRED_ENVS if env in os.environ}

    def get_runtime_env(self):
        runtime_env = {self.plugin_package_manager: [], 'working_dir': self.plugin_url}

        if self.requirements:
            runtime_env[self.plugin_package_manager] += self.requirements

        if self.debug:
            runtime_env[self.plugin_package_manager] += self.debug_modules

        # 맨 마지막에 진행되어야 함
        runtime_env['env_vars'] = self.envs

        if self.debug:
            pprint(runtime_env)
        return runtime_env

    def validate_params(self):
        if self.params_model:
            try:
                self.params_model.model_validate(self.params, context={'action': self})
            except ValidationError as e:
                raise ActionError({'params': pydantic_to_drf_error(e)})

    def run_action(self):
        self.validate_params()

        if self.direct:
            if self.method == RunMethod.RESTAPI:
                return self.start_by_restapi()
            else:
                result = self.start()
                if self.job_id:
                    self.post_action_by_job(result)
                return result
        return getattr(self, f'start_by_{self.method.value}')()

    def start(self):
        """Start the action.

        TODO: Specify the return type of start method for overrided methods.
        """
        if self.method == RunMethod.JOB:
            return self.entrypoint(self.run, **self.params)
        return self.entrypoint(**self.params)

    def start_by_task(self):
        """Ray Task based execution.

        * A task method that simply executes the entrypoint without job management functionality.
        """
        import ray
        from ray.exceptions import RayTaskError

        @ray.remote(runtime_env=self.get_runtime_env())
        def run_task(category, action, *args, **kwargs):
            from synapse_sdk.plugins.utils import get_action_class

            action = get_action_class(category, action)(*args, **kwargs)
            return action.run_action()

        init_signature = inspect.signature(self.__class__.__init__)

        args = []
        kwargs = {}

        for param in init_signature.parameters.values():
            if param.name == 'self':
                continue
            if param.default == param.empty:
                args.append(getattr(self, param.name))
            else:
                kwargs[param.name] = getattr(self, param.name)

        kwargs['direct'] = True
        try:
            self.ray_init()
            return ray.get(run_task.remote(self.category.value, self.name, *args, **kwargs))
        except RayTaskError as e:
            raise ActionError(e.cause)

    def start_by_job(self):
        """Ray Job based execution.

        * Executes the entrypoint with Ray job. Ray job manages the entrypoint execution and stores the results.
        """
        main_options = []
        options = ['run', '--direct']
        arguments = [self.name, f'{json.dumps(json.dumps(self.params))}']

        if self.debug:
            main_options.append('--debug')

        if self.job_id:
            options.append(f'--job-id={self.job_id}')

        cmd = ' '.join(main_options + options + arguments)

        client = self.get_job_client()
        return client.submit_job(
            submission_id=self.job_id,
            entrypoint=f'synapse plugin {cmd}',
            runtime_env=self.get_runtime_env(),
        )

    def start_by_restapi(self):
        """Ray Serve based execution.

        * This method executes a Fastapi endpoint defined within the Plugin.
        """
        path = self.params.pop('path', '')
        method = self.params.pop('method')

        url = self.plugin_release.get_serve_url(self.envs['RAY_SERVE_ADDRESS'], path)
        try:
            response = getattr(requests, method)(url, **self.params)
            try:
                response_body = response.json()
            except ValueError:
                response_body = response.text
            if response.ok:
                return response_body
            else:
                raise ActionError({'status': response.status_code, 'reason': response.reason, 'message': response_body})
        except requests.ConnectionError:
            raise ActionError('Unable to connect to serve application')

    def post_action_by_job(self, result):
        job_client = self.get_job_client()
        logs = job_client.get_job_logs(self.job_id).split('\n')
        self.client.update_job(self.job_id, data={'result': result or {}, 'console_logs': logs})

    def get_job_client(self):
        from ray.dashboard.modules.job.sdk import JobSubmissionClient

        return JobSubmissionClient(address=self.envs.get('RAY_DASHBOARD_URL'))

    def ray_init(self):
        import ray

        if not ray.is_initialized():
            ray.init(address=self.envs['RAY_ADDRESS'], ignore_reinit_error=True)



import asyncio
import base64
import hashlib
import json
import mimetypes
import operator
import zipfile
from functools import reduce
from pathlib import Path

import aiohttp
import requests
import yaml

from synapse_sdk.utils.network import clean_url
from synapse_sdk.utils.string import hash_text


def download_file(url, path_download, name=None, coerce=None, use_cached=True):
    chunk_size = 1024 * 1024 * 50
    cleaned_url = clean_url(url)  # remove query params and fragment

    if name:
        use_cached = False
    else:
        name = hash_text(cleaned_url)

    name += Path(cleaned_url).suffix

    path = Path(path_download) / name

    if not use_cached or not path.is_file():
        response = requests.get(url, allow_redirects=True, stream=True)
        response.raise_for_status()

        with path.open('wb') as file:
            for chunk in response.iter_content(chunk_size=chunk_size):
                file.write(chunk)

    if coerce:
        path = coerce(path)

    return path


def files_url_to_path(files, coerce=None, file_field=None):
    path_download = get_temp_path('media')
    path_download.mkdir(parents=True, exist_ok=True)
    if file_field:
        files[file_field] = download_file(files[file_field], path_download, coerce=coerce)
    else:
        for file_name in files:
            if isinstance(files[file_name], str):
                files[file_name] = download_file(files[file_name], path_download, coerce=coerce)
            else:
                files[file_name]['path'] = download_file(files[file_name].pop('url'), path_download, coerce=coerce)


def files_url_to_path_from_objs(objs, files_fields, coerce=None, is_list=False, is_async=False):
    if is_async:
        asyncio.run(afiles_url_to_path_from_objs(objs, files_fields, coerce=coerce, is_list=is_list))
    else:
        if not is_list:
            objs = [objs]

        for obj in objs:
            for files_field in files_fields:
                try:
                    files = reduce(operator.getitem, files_field.split('.'), obj)
                    if isinstance(files, str):
                        files_url_to_path(obj, coerce=coerce, file_field=files_field)
                    else:
                        files_url_to_path(files, coerce=coerce)
                except KeyError:
                    pass


async def adownload_file(url, path_download, name=None, coerce=None, use_cached=True):
    chunk_size = 1024 * 1024 * 50
    cleaned_url = clean_url(url)  # remove query params and fragment

    if name:
        use_cached = False
    else:
        name = hash_text(cleaned_url)

    name += Path(cleaned_url).suffix

    path = Path(path_download) / name

    if not use_cached or not path.is_file():
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                with path.open('wb') as file:
                    while chunk := await response.content.read(chunk_size):
                        file.write(chunk)

    if coerce:
        path = coerce(path)

    return path


async def afiles_url_to_path(files, coerce=None):
    path_download = get_temp_path('media')
    path_download.mkdir(parents=True, exist_ok=True)
    for file_name in files:
        if isinstance(files[file_name], str):
            files[file_name] = await adownload_file(files[file_name], path_download, coerce=coerce)
        else:
            files[file_name]['path'] = await adownload_file(files[file_name].pop('url'), path_download, coerce=coerce)


async def afiles_url_to_path_from_objs(objs, files_fields, coerce=None, is_list=False):
    if not is_list:
        objs = [objs]

    tasks = []

    for obj in objs:
        for files_field in files_fields:
            try:
                files = reduce(operator.getitem, files_field.split('.'), obj)
                tasks.append(afiles_url_to_path(files, coerce=coerce))
            except KeyError:
                pass

    await asyncio.gather(*tasks)


def get_dict_from_file(file_path):
    if isinstance(file_path, str):
        file_path = Path(file_path)

    with open(file_path) as f:
        if file_path.suffix == '.yaml':
            return yaml.safe_load(f)
        else:
            return json.load(f)


def calculate_checksum(file_path, prefix=''):
    md5_hash = hashlib.md5()
    with open(file_path, 'rb') as f:
        for byte_block in iter(lambda: f.read(4096), b''):
            md5_hash.update(byte_block)
    checksum = md5_hash.hexdigest()
    if prefix:
        return f'dev-{checksum}'
    return checksum


def archive(input_path, output_path):
    input_path = Path(input_path)
    output_path = Path(output_path)

    with zipfile.ZipFile(output_path, mode='w', compression=zipfile.ZIP_DEFLATED) as zipf:
        if input_path.is_file():
            zipf.write(input_path, input_path.name)
        else:
            for file_path in input_path.rglob('*'):
                if file_path.is_file():  # Only add files, skip directories
                    arcname = file_path.relative_to(input_path.parent)
                    zipf.write(file_path, arcname)


def unarchive(file_path, output_path):
    """
    Unarchives a ZIP file to a given directory.

    Parameters:
        file_path (str | Path): The path to the ZIP file.
        output_path (str): The directory where the files will be extracted.
    """
    output_path = Path(output_path)
    output_path.mkdir(parents=True, exist_ok=True)

    with zipfile.ZipFile(str(file_path), 'r') as zip_ref:
        zip_ref.extractall(output_path)


def get_temp_path(sub_path=None):
    path = Path('/tmp/datamaker')
    if sub_path:
        path = path / sub_path
    return path


def convert_file_to_base64(file_path):
    """
    Convert a file to base64 using pathlib.

    Args:
        file_path (str): Path to the file to convert

    Returns:
        str: Base64 encoded string of the file contents
    """
    # Convert string path to Path object
    path = Path(file_path)

    try:
        # Read binary content of the file
        binary_content = path.read_bytes()

        # Convert to base64
        base64_encoded = base64.b64encode(binary_content).decode('utf-8')

        # Get the MIME type of the file
        mime_type, _ = mimetypes.guess_type(path)
        assert mime_type is not None, 'MIME type cannot be guessed'

        # Convert bytes to string for readable output
        return f'data:{mime_type};base64,{base64_encoded}'

    except FileNotFoundError:
        raise FileNotFoundError(f'File not found: {file_path}')
    except Exception as e:
        raise Exception(f'Error converting file to base64: {str(e)}')



import time


def train(run, dataset, hyperparameter, checkpoint=None):
    count_epochs = hyperparameter['epochs']
    run.set_progress(0, count_epochs, category='train')
    for i in range(1, count_epochs + 1):
        time.sleep(0.5)
        loss = float(round((count_epochs - i) / count_epochs, 2))
        miou = 1 - loss
        run.log_metric('epoch', i, loss=loss, miou=miou)
        run.set_progress(i, count_epochs, category='train')
    return './'



from abc import ABC, abstractmethod
from typing import Any, Literal

from pydantic import BaseModel, field_validator
from pydantic_core import PydanticCustomError

from synapse_sdk.clients.exceptions import ClientError
from synapse_sdk.i18n import gettext as _
from synapse_sdk.plugins.categories.base import Action
from synapse_sdk.plugins.categories.decorators import register_action
from synapse_sdk.plugins.enums import PluginCategory, RunMethod
from synapse_sdk.utils.storage import get_pathlib


class ExportTargetHandler(ABC):
    """
    Abstract base class for handling export targets.

    This class defines the blueprint for export target handlers, requiring the implementation
    of methods to validate filters, retrieve results, and process collections of results.
    """

    @abstractmethod
    def validate_filter(self, value: dict, client: Any):
        """
        Validate filter query params to request original data from api.

        Args:
            value (dict): The filter criteria to validate.
            client (Any): The client used to validate the filter.

        Raises:
            PydanticCustomError: If the filter criteria are invalid.

        Returns:
            dict: The validated filter criteria.
        """
        pass

    @abstractmethod
    def get_results(self, client: Any, filters: dict):
        """
        Retrieve original data from target sources.

        Args:
            client (Any): The client used to retrieve the results.
            filters (dict): The filter criteria to apply.

        Returns:
            tuple: A tuple containing the results and the total count of results.
        """
        pass

    @abstractmethod
    def get_export_item(self, results):
        """
        Providing elements to build export data.

        Args:
            results (list): The results to process.

        Yields:
            generator: A generator that yields processed data items.
        """
        pass


class AssignmentExportTargetHandler(ExportTargetHandler):
    def validate_filter(self, value: dict, client: Any):
        if 'project' not in value:
            raise PydanticCustomError('missing_field', _('Project is required for Assignment.'))
        try:
            client.list_assignments(params=value)
        except ClientError:
            raise PydanticCustomError('client_error', _('Unable to get Assignment.'))
        return value

    def get_results(self, client: Any, filters: dict):
        return client.list_assignments(params=filters, list_all=True)

    def get_export_item(self, results):
        for result in results:
            yield {
                'data': result['data'],
                'files': result['file'],
                'id': result['id'],
            }


class GroundTruthExportTargetHandler(ExportTargetHandler):
    def validate_filter(self, value: dict, client: Any):
        if 'ground_truth_dataset_version' not in value:
            raise PydanticCustomError('missing_field', _('Ground Truth dataset version is required.'))
        try:
            client.get_ground_truth_version(value['ground_truth_dataset_version'])
        except ClientError:
            raise PydanticCustomError('client_error', _('Unable to get Ground Truth dataset version.'))
        return value

    def get_results(self, client: Any, filters: dict):
        filters['ground_truth_dataset_versions'] = filters.pop('ground_truth_dataset_version')
        return client.list_ground_truth_events(params=filters, list_all=True)

    def get_export_item(self, results):
        for result in results:
            files_key = next(iter(result['data_unit']['files']))
            yield {
                'data': result['data'],
                'files': result['data_unit']['files'][files_key],
                'id': result['ground_truth'],
            }


class TaskExportTargetHandler(ExportTargetHandler):
    def validate_filter(self, value: dict, client: Any):
        if 'project' not in value:
            raise PydanticCustomError('missing_field', _('Project is required for Task.'))
        try:
            client.list_tasks(params=value)
        except ClientError:
            raise PydanticCustomError('client_error', _('Unable to get Task.'))
        return value

    def get_results(self, client: Any, filters: dict):
        filters['expand'] = 'data_unit'
        return client.list_tasks(params=filters, list_all=True)

    def get_export_item(self, results):
        for result in results:
            files_key = next(iter(result['data_unit']['files']))
            yield {
                'data': result['data'],
                'files': result['data_unit']['files'][files_key],
                'id': result['id'],
            }


class TargetHandlerFactory:
    @staticmethod
    def get_handler(target: str) -> ExportTargetHandler:
        if target == 'assignment':
            return AssignmentExportTargetHandler()
        elif target == 'ground_truth':
            return GroundTruthExportTargetHandler()
        elif target == 'task':
            return TaskExportTargetHandler()
        else:
            raise ValueError(f'Unknown target: {target}')


class ExportParams(BaseModel):
    """
    Parameters for the export action.

    Attributes:
        storage (int): The storage ID to save the exported data.
        save_original_file (bool): Whether to save the original file.
        path (str): The path to save the exported data.
        target (str): The target source to export data from. (ex. ground_truth, assignment, task)
        filter (dict): The filter criteria to apply.
    """

    storage: int
    save_original_file: bool = True
    path: str
    target: Literal['assignment', 'ground_truth', 'task']
    filter: dict

    @field_validator('storage')
    @staticmethod
    def check_storage_exists(value, info):
        action = info.context['action']
        client = action.client
        try:
            client.get_storage(value)
        except ClientError:
            raise PydanticCustomError('client_error', _('Unable to get storage from Synapse backend.'))
        return value

    @field_validator('filter')
    @staticmethod
    def check_filter_by_target(value, info):
        action = info.context['action']
        client = action.client
        target = action.params['target']
        handler = TargetHandlerFactory.get_handler(target)
        return handler.validate_filter(value, client)


@register_action
class ExportAction(Action):
    name = 'export'
    category = PluginCategory.EXPORT
    method = RunMethod.JOB
    params_model = ExportParams
    progress_categories = {
        'dataset_conversion': {
            'proportion': 100,
        }
    }

    def get_filtered_results(self, filters, handler):
        """Get filtered target results."""
        try:
            result_list = handler.get_results(self.client, filters)
            results = result_list[0]
            count = result_list[1]
        except ClientError:
            raise PydanticCustomError('client_error', _('Unable to get Ground Truth dataset.'))
        return results, count

    def start(self):
        filters = {'expand': 'data', **self.params['filter']}
        target = self.params['target']
        handler = TargetHandlerFactory.get_handler(target)

        self.params['results'], self.params['count'] = self.get_filtered_results(filters, handler)
        export_items = handler.get_export_item(self.params['results'])

        storage = self.client.get_storage(self.params['storage'])
        pathlib_cwd = get_pathlib(storage, self.params['path'])
        return self.entrypoint(self.run, export_items, pathlib_cwd, **self.params)



from upath import UPath

from synapse_sdk.utils.storage.providers import BaseStorage


class S3Storage(BaseStorage):
    ENDPOINT_URL = 'https://s3.amazonaws.com'
    DEFAULT_REGION = 'us-east-1'

    def __init__(self, url):
        super().__init__(url)

        self.upath = self._get_upath()

    def _get_upath(self):
        upath_kwargs = {
            'key': self.query_params['access_key'],
            'secret': self.query_params['secret_key'],
            'client_kwargs': {'region_name': self.query_params.get('region_name')},
        }

        if self.query_params.get('endpoint_url'):
            upath_kwargs['endpoint_url'] = self.query_params['endpoint_url']

        return UPath(
            f's3://{self.query_params["bucket_name"]}',
            **upath_kwargs,
        )

    def upload(self, source, target):
        with open(source, 'rb') as file:
            self.upath.write_text(file.read(), target)

        return self.get_url(target)

    def exists(self, target):
        return self.upath.exists(target)

    def get_url(self, target):
        return str(self.upath.joinuri(target))

    def get_pathlib(self, path):
        return self.upath.joinuri(path)



from synapse_sdk.clients.base import BaseClient
from synapse_sdk.clients.utils import get_default_url_conversion


class AnnotationClientMixin(BaseClient):
    def get_project(self, pk):
        path = f'projects/{pk}/'
        return self._get(path)

    def get_task_tag(self, pk):
        path = f'task_tags/{pk}/'
        return self._get(path)

    def list_task_tags(self, params):
        path = 'task_tags/'
        return self._list(path, params=params)

    def list_tasks(self, params=None, url_conversion=None, list_all=False):
        path = 'tasks/'
        url_conversion = get_default_url_conversion(url_conversion, files_fields=['files'])
        return self._list(path, params=params, url_conversion=url_conversion, list_all=list_all)

    def create_tasks(self, data):
        path = 'tasks/'
        return self._post(path, data=data)

    def set_tags_tasks(self, data, params=None):
        path = 'tasks/set_tags/'
        return self._post(path, data=data, params=params)



from synapse_sdk.clients.base import BaseClient
from synapse_sdk.clients.exceptions import ClientError


class RayClientMixin(BaseClient):
    def get_job(self, pk):
        path = f'jobs/{pk}/'
        return self._get(path)

    def list_jobs(self):
        path = 'jobs/'
        return self._get(path)

    def list_job_logs(self, pk):
        path = f'jobs/{pk}/logs/'
        return self._get(path)

    def tail_job_logs(self, pk):
        if self.long_poll_handler:
            raise ClientError(400, '"tail_job_logs" does not support long polling')

        path = f'jobs/{pk}/tail_logs/'

        url = self._get_url(path)
        headers = self._get_headers()

        response = self.requests_session.get(url, headers=headers, stream=True)
        for line in response.iter_lines(decode_unicode=True):
            if line:
                yield f'{line}\n'

    def get_node(self, pk):
        path = f'nodes/{pk}/'
        return self._get(path)

    def list_nodes(self):
        path = 'nodes/'
        return self._get(path)

    def get_task(self, pk):
        path = f'tasks/{pk}/'
        return self._get(path)

    def list_tasks(self):
        path = 'tasks/'
        return self._get(path)

    def get_serve_application(self, pk):
        path = f'serve_applications/{pk}/'
        return self._get(path)

    def list_serve_applications(self):
        path = 'serve_applications/'
        return self._get(path)

    def delete_serve_application(self, pk):
        path = f'serve_applications/{pk}/'
        return self._delete(path)



import json
from pathlib import Path

from synapse_sdk.plugins.utils import get_plugin_categories
from synapse_sdk.utils.file import get_dict_from_file


def update_config(config):
    config['category'] = get_plugin_categories()
    return config


def main():
    cookiecutter_path = Path('cookiecutter.json')
    config = get_dict_from_file(cookiecutter_path)
    config = update_config(config)
    cookiecutter_path.write_text(json.dumps(config, indent=4, ensure_ascii=False), encoding='utf-8')


if __name__ == '__main__':
    main()



import click

from .utils import delete_alias_file, load_alias


@click.command()
@click.argument('name', required=True)
def delete(name):
    """Delete an alias."""
    alias_data = load_alias(name)
    if not alias_data:
        click.echo(f"Alias '{name}' does not exist.")
        return
    delete_alias_file(name)
    click.echo(f"Alias '{name}' deleted.")



from pathlib import Path
from urllib.parse import urlparse

from synapse_sdk.i18n import gettext as _
from synapse_sdk.utils.storage.registry import STORAGE_PROVIDERS


def get_storage(connection_param: str | dict):
    """Get storage class with connection param.

    Args:
        connection_param (str | dict): The connection param for the Storage provider.

    Returns:
        BaseStorage: The storage class object with connection param.
    """
    storage_scheme = None
    if isinstance(connection_param, dict):
        storage_scheme = connection_param['provider']
    else:
        storage_scheme = urlparse(connection_param).scheme

    assert storage_scheme in STORAGE_PROVIDERS.keys(), _('Storage provider not supported.')
    return STORAGE_PROVIDERS[storage_scheme](connection_param)


def get_pathlib(storage_config: str | dict, path_root: str) -> Path:
    """Get pathlib object with synapse-backend storage config.

    Args:
        storage_config (str | dict): The storage config by synapse-backend storage api.
        path_root (str): The path root.

    Returns:
        pathlib.Path: The pathlib object.
    """
    storage_class = get_storage(storage_config)
    return storage_class.get_pathlib(path_root)



from pathlib import Path

import click
from cookiecutter.main import cookiecutter


@click.command()
def create():
    project_root = Path(__file__).parents[2]
    cookiecutter(str(project_root / 'plugins/templates'))



from pathlib import Path
from typing import List


class Uploader:
    """Plugin upload action class.

    * Organize, upload, setup task, generate ground truths for the uploaded data.
    """

    def __init__(self, run, path: Path, *args, **kwargs):
        """Initialize the plugin upload action class.

        Args:
            run: Plugin run object.
            path: pathlib object by upload target destination path.
        """
        self.run = run
        self.path = path

    def handle_upload_files(self) -> List:
        """Handle upload files.

        * Organize data according to collection file specification structure.
        * Structure files according to the file specification of the target collection.

        Returns:
            List: List of dictionaries containing 'files' and 'meta'.

        Examples:
            [
              {
          "files": {
              'image_1': image_1_pathlib_object,
              'image_2': image_2_pathlib_object,
              'meta_1': meta_1_pathlib_object,
          },
          "meta": {
            "key": "value"
          }
              }
            ]
        """
        return []



from dataclasses import dataclass
from typing import Union


@dataclass
class EnvVar:
    name: str
    type = str
    default: Union[str, None] = None
    show_default: bool = True


ENV_VARS = {
    'RAY_ADDRESS': EnvVar('RAY_ADDRESS'),
    'RAY_DASHBOARD_ADDRESS': EnvVar('RAY_DASHBOARD_ADDRESS'),
    'RAY_SERVE_ADDRESS': EnvVar('RAY_SERVE_ADDRESS'),
    'SYNAPSE_PLUGIN_STORAGE': EnvVar('SYNAPSE_PLUGIN_STORAGE', default='', show_default=False),
    'SYNAPSE_DEBUG_PLUGIN_PATH': EnvVar('SYNAPSE_DEBUG_PLUGIN_PATH', default='', show_default=False),
    'SYNAPSE_DEBUG_MODULES': EnvVar('SYNAPSE_DEBUG_MODULES', default='', show_default=False),
    'SYNAPSE_PLUGIN_PUBLISH_HOST': EnvVar('SYNAPSE_PLUGIN_PUBLISH_HOST'),
    'SYNAPSE_PLUGIN_PUBLISH_USER_TOKEN': EnvVar('SYNAPSE_PLUGIN_PUBLISH_USER_TOKEN'),
    'SYNAPSE_PLUGIN_PUBLISH_TENANT': EnvVar('SYNAPSE_PLUGIN_PUBLISH_TENANT'),
    'SYNAPSE_PLUGIN_RUN_HOST': EnvVar('SYNAPSE_PLUGIN_RUN_HOST', default='SYNAPSE_PLUGIN_PUBLISH_HOST'),
    'SYNAPSE_PLUGIN_RUN_AGENT': EnvVar('SYNAPSE_PLUGIN_RUN_AGENT', default='', show_default=False),
    'SYNAPSE_PLUGIN_RUN_AGENT_TOKEN': EnvVar('SYNAPSE_PLUGIN_RUN_AGENT_TOKEN'),
    'SYNAPSE_PLUGIN_RUN_AGENT_HOST': EnvVar('SYNAPSE_PLUGIN_RUN_AGENT_HOST'),
    'SYNAPSE_PLUGIN_RUN_USER_TOKEN': EnvVar(
        'SYNAPSE_PLUGIN_RUN_USER_TOKEN', default='SYNAPSE_PLUGIN_PUBLISH_USER_TOKEN'
    ),
    'SYNAPSE_PLUGIN_RUN_TENANT': EnvVar('SYNAPSE_PLUGIN_RUN_TENANT', default='SYNAPSE_PLUGIN_PUBLISH_TENANT'),
}
